#include <ISA.h>
#include <due_can.h>
#include <DueTimer.h>
#include <DueFlashStorage.h>
#include <efc.h>
#include <flash_efc.h>
#include "globals.h"
#include "chademo.h"
/*
  Notes on what needs to be done:
  - Timing analysis showed that the USB, CANBUS, and BT routines take up entirely too much time. They can delay processing by
   almost 100ms! Minor change for test.

  - Investigate what changes are necessary to support the Cortex M0 processor in the Arduino Zero

  - Interrupt driven CAN has a tendency to lock up. It has been disabled for now - It locks up even if the JLD is not sending anything
  but it seems to be able to actually send as much as you want. Only interrupt driven reception seems to make things die.

  Note about timing related code: The function millis() returns a 32 bit integer that specifies the # of milliseconds since last start up.
  That's all well and good but 4 billion milliseconds is a little less than 50 days. One might ask "so?" well, this could potentially run
  indefinitely in some vehicles and so would suffer rollover every 50 days. When this happens improper code will become stupid. So, try
  to implement any timing code like this:
  if ((millis() - someTimeStamp) >= SomeInterval) do stuff

  Such a code block will do the proper thing so long as all variables used are unsigned long / uint32_t variables.

*/

//#define DEBUG_TIMING	//if this is defined you'll get time related debugging messages
//#define WIFI_ENABLED

template<class T> inline Print &operator <<(Print &obj, T arg) {
  obj.print(arg);  //Sets up serial streaming Serial<<someshit;
  return obj;
}



//These have been moved to eeprom. After initial compile the values will be read from EEPROM.
//These thus set the default value to write to eeprom upon first start up
#define MAX_CHARGE_V	158
#define MAX_CHARGE_A	130
#define TARGET_CHARGE_V	160
#define MIN_CHARGE_A	20
#define INITIAL_SOC 100
#define CAPACITY 180
#define EEPROM_BASE 255
#define PRINT_USB 1
#define PRINT_BT 2
#define STATE_USB 1
#define STATE_BT 2

/**
   Output modes for OUT0 and OUT1

   In each case an Hysteresis applies after crossing the threshold to reset the output.
*/
#define OUTMODE_OFF 0       //OFF
#define OUTMODE_V_LT 1      //Voltage less than threshold
#define OUTMODE_V_GT 2      //Voltage greater than threshold
#define OUTMODE_A_LT 3      //Current less than threshold
#define OUTMODE_A_GT 4      //Current greater than threshold
#define OUTMODE_S_LT 5      //State of charge less than threshold
#define OUTMODE_S_GT 6      //State of charge greater than threshold
#define OUTMODE_V_LT_AH 7   //Same as OUTMODE_V_LT but also triggers if the AH > Capacity >> special case.


const unsigned long Interval = 10;
unsigned long Time = 0;
unsigned long ChargeTimeRefSecs = 0; // reference time for charging.
unsigned long PreviousMillis = 0;
unsigned long CurrentMillis = 0;
float Voltage = 0;
float Current = 0;
float Power = 0;
float lastSavedAH = 0;
float Debug_Amps = 0;
int Count = 0;
byte Command = 0; // "z" will reset the AmpHours and KiloWattHours counters
volatile uint8_t bStartConversion = 0;
volatile uint8_t bGetTemperature = 0;
volatile uint8_t timerIntCounter = 0;
volatile uint8_t timerFastCounter  = 0;
volatile uint8_t timerChademoCounter = 0;
volatile uint8_t sensorReadPosition = 255;
uint8_t tempSensorCount = 0;
volatile uint8_t debugTick = 0;
volatile uint8_t wifiTick = 0;


uint16_t uint16Val;
uint16_t uint16Val2;
uint8_t uint8Val;
uint8_t print8Val;
uint8_t state8Val;
byte limpMode;

String strVal;
String cmdStr;

EESettings settings;
DueFlashStorage dueFlashStorage;
#define EEPROM_VALID	0xEE

ISA Sensor;


void timer2Int()
{
  timerFastCounter++;
  timerChademoCounter++;
  if (timerChademoCounter >= 3)
  {
    timerChademoCounter = 0;
    if (chademo.bChademoMode  && chademo.bChademoSendRequests) chademo.bChademoRequest = 1;
  }

  if (timerFastCounter == 8)
  {
    debugTick = 1;
    wifiTick = 1;
    timerFastCounter = 0;
    timerIntCounter++;
    if (timerIntCounter < 10)
    {
      bGetTemperature = 1;
      sensorReadPosition++;
    }
    if (timerIntCounter == 10)
    {
      bStartConversion = 1;
      sensorReadPosition = 255;
    }
    if (timerIntCounter == 18)
    {
      timerIntCounter = 0;
    }
  }
}

void setup()
{
  //first thing configure the I/O pins and set them to a sane state
  pinMode(IN0, INPUT);
  pinMode(IN1, INPUT);
  pinMode(OUT0, OUTPUT);
  pinMode(OUT1, OUTPUT);
  digitalWrite(OUT0, LOW);
  digitalWrite(OUT1, LOW);

  SerialUSB.begin(115200);//normal port
  Serial2.begin(19200);//wifi port

  Can0.begin(CAN_BPS_500K);
  Can0.watchFor();

  Sensor.begin(0, 500); //Start ISA object on CAN 0 at 500 kbps

  byte* b = dueFlashStorage.readAddress(4); // byte array which is read from flash at adress 4
  memcpy(&settings, b, sizeof(EESettings)); // copy byte array to temporary struct

  if (settings.valid != EEPROM_VALID) //not proper version so reset to defaults
  {
    settings.packSizeKWH = 33.0; //just a random guess. Maybe it should default to zero though?
    settings.valid = EEPROM_VALID;
    settings.ampHours = 0.0;
    settings.kiloWattHours = 0.0; // Is empty kiloWattHours count up
    settings.maxChargeAmperage = MAX_CHARGE_A;
    settings.maxChargeVoltage = MAX_CHARGE_V;
    settings.targetChargeVoltage = TARGET_CHARGE_V;
    settings.minChargeAmperage = MIN_CHARGE_A;
    settings.SOC = INITIAL_SOC;
    settings.capacity = CAPACITY;
    settings.debuggingLevel = 1;
    settings.currentOffset = 0;
    Save();
  }

  print8Val = PRINT_USB;
  state8Val = 0;
  if (settings.debuggingLevel > 1)
    settings.debuggingLevel = 0; //locked in to max debugging for now.

  limpMode = 0;

  Timer2.attachInterrupt(timer2Int);
  Timer2.start(25000); // Calls every 25ms

  updateTargetAV();
}


void loop()
{
  uint8_t pos;
  CurrentMillis = millis();
  uint8_t len;
  CAN_FRAME inFrame;
  float tempReading;

#ifdef DEBUG_TIMING
  if (debugTick == 1)
  {
    debugTick = 0;
    SerialUSB.println(millis());
  }
#endif

#ifdef WIFI_ENABLED
  if (wifiTick == 1)
  {
    handle_wifi();
    wifiTick = 0;
  }
#endif

  chademo.loop();

  if (CurrentMillis - PreviousMillis >= Interval)
  {
    Time = CurrentMillis - PreviousMillis;
    PreviousMillis = CurrentMillis;

    Count++;
    Voltage = Sensor.Voltage;
    Current = Sensor.Amperes;
    settings.ampHours += Current * (float)Time / 1000.0 / 3600.0;
    Power = Voltage * Current / 1000.0;

    //Count down kiloWattHours when drawing current.
    //Count up when charging (Current/Power is negative)
    settings.kiloWattHours = Sensor.KWH;
    settings.SOC = ((settings.capacity - settings.ampHours) / settings.capacity) * 100;

    if (settings.ampHours > settings.capacity && Current > 50)
      limpMode = 1;

    chademo.doProcessing();

    if (Count >= 50)
    {
      Count = 0;

      USB();
      WIFI();

      if (print8Val > 0)
        printSettings();
      else
        outputState();

      if (chademo.bChademoMode)
      {
        if (settings.debuggingLevel > 0) {
          SerialUSB.print(F("Chademo Mode: "));
          SerialUSB.println(chademo.getState());
        }
      }
      else
      {
        checkChargingState(); // looking for reset of charging.
      }
    }
  }
  if (Can0.available() > 0) {
    CAN.read(inFrame);
    chademo.handleCANFrame(inFrame);
  }
}

void updateTargetAV()
{
  chademo.setTargetAmperage(settings.maxChargeAmperage);
  chademo.setTargetVoltage(settings.targetChargeVoltage);
}

void Save()
{
  noInterrupts();
  byte b2[sizeof(EESettings)]; // create byte array to store the struct
  memcpy(b2, &settings, sizeof(EESettings)); // copy the struct to the byte array
  dueFlashStorage.write(4, b2, sizeof(EESettings)); // write byte array to flash
  lastSavedAH = settings.ampHours;
  SerialUSB.println (F("SAVED"));
  interrupts();
}

void RngErr() {
  SerialUSB.println(F("Range Error"));
}

void SetAH(float ah) {
  settings.ampHours = ah;
  float val = ((settings.capacity - settings.ampHours) / settings.capacity);
  settings.kiloWattHours = val * settings.packSizeKWH;
  settings.SOC = val * 100;
}

void ParseCommand(int bt) {
  float fltVal = 0;
  if (cmdStr == "AH") {
    fltVal = bt > 0 ? Serial2.parseFloat() : SerialUSB.parseFloat();
    if (fltVal >= 0 && fltVal < 1000) {
      SetAH(fltVal);
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "V") {
    uint16Val = bt > 0 ? Serial2.parseInt() : SerialUSB.parseInt();
    if (uint16Val > 0 && uint16Val < 1000) {
      settings.targetChargeVoltage = uint16Val;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "VMAX") {
    uint16Val = bt > 0 ? Serial2.parseInt() : SerialUSB.parseInt();
    if (uint16Val > 0 && uint16Val < 1000) {
      settings.maxChargeVoltage = uint16Val;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "AMIN") {
    uint8Val = bt > 0 ? Serial2.parseInt() : SerialUSB.parseInt();
    if (uint8Val > 0 && uint8Val < 256) {
      settings.minChargeAmperage = uint8Val;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "AMAX") {
    uint8Val = bt > 0 ? Serial2.parseInt() : SerialUSB.parseInt();
    if (uint8Val > 0 && uint8Val < 256) {
      settings.maxChargeAmperage = uint8Val;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "CAPAH") {
    uint8Val = bt > 0 ? Serial2.parseInt() : SerialUSB.parseInt();
    if (uint8Val > 0 && uint8Val < 256) {
      settings.capacity = uint8Val;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "CAPKWH") {
    fltVal = bt > 0 ? Serial2.parseFloat() : SerialUSB.parseFloat();
    if (fltVal > 0 && fltVal < 100) {
      settings.packSizeKWH = fltVal;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "DBG") {
    uint8Val = bt > 0 ? Serial2.parseInt() : SerialUSB.parseInt();
    if (uint8Val >= 0 && uint8Val < 4) {
      settings.debuggingLevel = uint8Val;
      Save();
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    } else {
      RngErr();
    }
  }
  else if (cmdStr == "SAVE") {
    Save();
  }
  else if (cmdStr == "FULLRESET") {
    settings.valid = 0;
    Save();
  }
}


void checkChargingState() {
  //If the current is negative we are either charging or regenerating.
  //If we are charging we expect a relatively constant current decreasing over
  //a period of time. Ie. a small rate of change of current while charging, throughout the charging period.
  unsigned long CurrentSecs = CurrentMillis / 1000;
  if (Current < -2.0 && Current > -20 && Voltage >= settings.targetChargeVoltage) { // Charging or regenerating at less than 20 amps
    //while we are charging batteries at less than 20 amps we assume if this occurs for longer than 5 minutes
    //we are charging and not regenerating
    if (Current > -15 && (CurrentSecs - ChargeTimeRefSecs) > 30) { // have been between 0 and -15 amps for more than 5 minutes
      //At this point we assume that we have been charging for the last 5 minutes or more prior to hitting -4 amps
      ResetChargeState();
      ChargeTimeRefSecs = CurrentSecs;
    }
  } else // Normal driving conditions or we don't care. So we keep resetting the reference.
    ChargeTimeRefSecs = CurrentSecs;
}

void ResetChargeState() {
  SerialUSB.println(F("Reset Ah & Wh"));
  settings.ampHours = 0.0;
  settings.kiloWattHours = settings.packSizeKWH;
  Save();
  print8Val = PRINT_USB;
  SerialUSB.println(F("Done!!!"));
}

void outputState() {

    SerialUSB.print (F("JLD505: "));
    SerialUSB.print (Voltage, 3);
    SerialUSB.print (F("v "));
    SerialUSB.print (Current, 2);
    SerialUSB.print (F("A "));
    SerialUSB.print (settings.ampHours, 1);
    SerialUSB.print (F("Ah "));
    SerialUSB.print (Power, 1);
    SerialUSB.print (F("kW "));
    SerialUSB.print (settings.kiloWattHours, 1);
    SerialUSB.print (F("kWh "));
    SerialUSB.print (settings.SOC, 1);
    SerialUSB.print (F("% SOC "));
    SerialUSB.print (F("OUT0"));
    SerialUSB.print (digitalRead(OUT0) > 0 ? F(":1 ") : F(":0 "));
    SerialUSB.print (F("OUT1"));
    SerialUSB.print (digitalRead(OUT1) > 0 ? F(":1") : F(":0 "));
    SerialUSB.print (F("IN0"));
    SerialUSB.print (digitalRead(IN0) > 0 ? F(":1 ") : F(":0 "));
    SerialUSB.print (F("IN1"));
    SerialUSB.print (digitalRead(IN1) > 0 ? F(":1") : F(":0 "));
    SerialUSB.print (F("CHG T: "));
    SerialUSB.println (CurrentMillis / 1000 - ChargeTimeRefSecs);
}

void USB() {
  SerialCommand(0);
}

void WIFI() {
  SerialCommand(1);
}

void SerialCommand(int bt)
{
  int available = bt > 0 ? Serial2.available() : SerialUSB.available();
  if (available > 0)
  {
    Command = bt > 0 ? Serial2.peek() : SerialUSB.peek();
    if (Command == 'z')
    {
      ResetChargeState();
    }
    if (Command == 'p')
    {
      print8Val = bt > 0 ? PRINT_BT : PRINT_USB;
    }
    if (Command == 'u')
    {
      updateTargetAV();
    }

    cmdStr = bt > 0 ? Serial2.readStringUntil('=') : SerialUSB.readStringUntil('=');
    ParseCommand(bt);

    if (bt > 0)
      while (Serial2.available() > 0) Serial2.read();
    else
      while (SerialUSB.available() > 0) SerialUSB.read();
  }

}

void printSettings() {
    SerialUSB.println (F("-"));
    SerialUSB.print (F("AH: "));
    SerialUSB.println (settings.ampHours);
    SerialUSB.print (F("KWH: "));
    SerialUSB.println (settings.kiloWattHours);
    SerialUSB.print (F("AMIN "));
    SerialUSB.println (settings.minChargeAmperage, 1);
    SerialUSB.print (F("AMAX "));
    SerialUSB.println (settings.maxChargeAmperage, 1);
    SerialUSB.print (F("V "));
    SerialUSB.println (settings.targetChargeVoltage, 1);
    SerialUSB.print (F("VMAX "));
    SerialUSB.println (settings.maxChargeVoltage, 1);
    SerialUSB.print (F("CAPAH "));
    SerialUSB.println (settings.capacity, 1);
    SerialUSB.print (F("CAPKWH: "));
    SerialUSB.println (settings.packSizeKWH, 2);
    SerialUSB.print (F("COFF: "));
    SerialUSB.println (settings.currentOffset, 2);
    SerialUSB.print (F("DBG "));
    SerialUSB.println (settings.debuggingLevel, 1);
    SerialUSB.println (F("-"));
  print8Val = 0;
}

void timestamp()
{
  int milliseconds = (int) (millis() / 1) % 1000 ;
  int seconds = (int) (millis() / 1000) % 60 ;
  int minutes = (int) ((millis() / (1000 * 60)) % 60);
  int hours   = (int) ((millis() / (1000 * 60 * 60)) % 24);

  SerialUSB.print(F(" Time:"));
  SerialUSB.print(hours);
  SerialUSB.print(F(":"));
  SerialUSB.print(minutes);
  SerialUSB.print(F(":"));
  SerialUSB.print(seconds);
  SerialUSB.print(F("."));
  SerialUSB.println(milliseconds);
}
void handle_wifi() {
  /*

     Routine to send data to wifi on serial 2
    The information will be provided over serial to the esp8266 at 19200 baud 8n1 in the form :
    vxxx,ixxx,pxxx,mxxxx,nxxxx,oxxx,rxxx,qxxx* where :

    v=pack voltage (0-700Volts)
    i=current (0-1000Amps)
    p=power (0-300kw)
    m=chademo state
    n=KWH

     =end of string
    xxx=three digit integer for each parameter eg p100 = 100kw.
    updates will be every 100ms approx.

    v100,i200,p35,m3000,n4000*

    Chademo state enums
    STARTUP,
    SEND_INITIAL_PARAMS,
    WAIT_FOR_EVSE_PARAMS,
    SET_CHARGE_BEGIN,
    WAIT_FOR_BEGIN_CONFIRMATION,
    CLOSE_CONTACTORS,
    RUNNING,
    CEASE_CURRENT,
    WAIT_FOR_ZERO_CURRENT,
    OPEN_CONTACTOR,
    FAULTED,
    STOPPED,
    LIMBO
  */

  //Serial2.print("v100,i200,p35,m3000,n4000,o20,r30,q50*"); //test string

  digitalWrite(13, !digitalRead(13)); //blink led every time we fire this interrrupt.

  Serial2.print("v");//dc bus voltage
  Serial2.print(Sensor.Voltage);//voltage derived from ISA shunt
  Serial2.print(",i");//dc current
  Serial2.print(Sensor.Amperes);//current derived from ISA shunt
  Serial2.print(",p");//total motor power
  Serial2.print(Sensor.KW);//Power value derived from ISA Shunt
  Serial2.print(",m");//Chademo state
  Serial2.print(chademo.getState());
  Serial2.print(",n");//Rear motor rpm
  Serial2.print(Sensor.KWH);// change to kwh on wifi.
  Serial2.print("*");// end of data indicator
}
